# -*- coding: utf-8 -*-

import logging
import functools
import time
import falcon
import blinker
import gevent

from gevent.hub import get_hub

from .exception import (
    BasicUserException,
    Unauthorized,
)

hub = get_hub()
hub.NOT_ERROR = (*hub.NOT_ERROR, falcon.http_status.HTTPStatus)


class FunctionExecutor(object):

    def __init__(self, http_wrapper=None, sentry_client=None, auth_func=None, log_exclude_fields=None, identity_func=None):
        self.http_wrapper = http_wrapper if http_wrapper else lambda d, s, m, c: d
        self.sentry_client = sentry_client
        self.auth_func = auth_func if auth_func else lambda req, func: True
        self.log_exclude_fields = log_exclude_fields if log_exclude_fields is not None else {}
        self.identity_func = identity_func

    def _http_wrapper(self, data=None, status=200, message='Success', code=0, response=None):
        if response:
            response.status = getattr(falcon, 'HTTP_{}'.format(status))
        return self.http_wrapper(data, status, message, code)

    def _process(self, func, args, kwargs, request, response):
        start_sig = blinker.signal('BeforeFunctionExecute')
        start_sig.send(request)

        if not self.auth_func(request, func):
            raise Unauthorized

        return self._http_wrapper(data=func(*args, **kwargs))

    def _send_sentry_exc(self, request):
        if self.sentry_client:
            if request:
                self.sentry_clinet.http_context({
                    'url': request.url,
                    'query_string': request.query_string,
                    'method': request.method,
                    'headers': request.headers,
                })
            self.sentry_client.captureException()

    def _prepare_log(self, func_name, args, kwargs, execution_time, request):
        if self.identity_func:
            kwargs['unique_identity'] = self.identity_func(request)
        kwargs_list = []
        for k, v in kwargs.items():
            if k not in self.ignore_fields:
                if isinstance(v, str):
                    v = "'{}'".format(v)
                kwargs_list.append('{}={}'.format(k, v))
        args_str = [str(a) for a in args] + kwargs_list
        return '{}({}) response {}ms'.format(
            func_name, ', '.join(args_str), execution_time)

    def _process_exception_output(self, e, func_logger, request, response):
        if self.http_wrapper:
            if isinstance(e, BasicUserException):
                return self._http_wrapper(
                    status=e.identity.http_code,
                    code=e.identity.code,
                    message=e.identity.error_msg,
                    response=response,
                )
            else:
                self._send_sentry_exc(request)
                func_logger.exception(e)
                return self._http_wrapper(
                    status=500,
                    code=0,
                    message='Internal Server Error',
                    response=response,
                )
        else:
            raise

    def _finish_exec(self, duration, func_logger, args, kwargs, request, func):
        end_sig = blinker.signal('AfterFunctionExecute')
        end_sig.send(request)
        func_logger.info(self._prepare_log(func.__name__, args, kwargs, duration, request))

    def __call__(self, func):
        func_logger = logging.getLogger(func.__module__)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            request = kwargs.get('request')
            if request:
                del kwargs['request']
            response = kwargs.get('response')
            if response:
                del kwargs['response']
            try:
                return self._process(func, args, kwargs, request, response)
            except falcon.http_status.HTTPStatus:
                raise
            except Exception as e:
                return self._process_exception_output(e, func_logger, request, response)
            finally:
                execution_time = (time.time() - start) * 1000
                self._finish_exec(execution_time, func_logger, args, kwargs, request, func)
        return wrapper



# -*- coding: utf-8 -*-

import subprocess
import os
import argparse
from contextlib import suppress
from jinja2 import Template
from gunicorn.app.base import BaseApplication


def init(args):
    print('Gathering information for initial...')
    app_name = input('App name: ')
    author_name = input('Author name: ')
    default_app_addr = 'https://github.com/{author_name}/{app_name}'.format(author_name=author_name, app_name=app_name)
    app_addr = input('App Home address (default: {addr}): '.format(addr=default_app_addr)) or default_app_addr
    database_addr = input('Database address (default: localhost)') or 'localhost'
    database_port = input('Database port (default: 5432)') or 5432
    database_user = input('Database user (default: postgres)') or 'postgres'
    database_password = input('Database password (default: postgres)') or 'postgres'
    database_db = input('Database dbname (default: postgres)') or 'postgres'

    package_infos = locals()
    # TODO: render infos into template and generate static python files.

    template_path = os.path.join(os.path.dirname(__file__), 'templates')
    for current_dir, _, files in os.walk(template_path):
        target_dir = os.path.relpath(current_dir, template_path)
        with suppress(FileExistsError, IsADirectoryError):
            os.mkdir(target_dir)

        for file_ in files:
            with open(os.path.join(current_dir, file_), 'r') as f:
                content = f.read()
            rendered = Template(content).render(**package_infos)
            with open(os.path.join(target_dir, file_.rstrip('.jinja')), 'w') as f:
                f.write(rendered)
    os.rename('app', app_name)
    subprocess.call(['python', 'setup.py', 'develop'])


def dev_run(args):
    try:
        from gunicorn_config import import_wsgi
    except ImportError:
        raise ImportError('Must run inside a Makiki app and keep gunicorn_config file untouched.')

    class DevServer(BaseApplication):
        def __init__(self, bind, workers):
            self._bind = bind
            self._workers = workers
            super().__init__()

        def load_config(self):
            self.cfg.set('bind', self._bind)
            self.cfg.set('workers', self._workers)
            self.cfg.set('worker_class', 'gevent')
            self.cfg.set('reload', True)

        def load(self):
            return import_wsgi()

    DevServer(args.bind, args.workers).run()


def main_parser():
    root_parser = argparse.ArgumentParser()
    sub_parsers = root_parser.add_subparsers()
    sub_parsers.required = True
    sub_parsers.dest = 'command'

    init_parser = sub_parsers.add_parser('init')
    init_parser.set_defaults(func=init)

    dev_run_parser = sub_parsers.add_parser('dev-run')
    dev_run_parser.add_argument('--bind', default='0.0.0.0:8000', help='default: 0.0.0.0:8000')
    dev_run_parser.add_argument('--workers', default=1, help='default: 1')
    dev_run_parser.set_defaults(func=dev_run)

    args = root_parser.parse_args()
    args.func(args)




import psycogreen.gevent

import functools
import json
import hug.interface

import hug._empty as empty

from cgi import parse_header
from collections import OrderedDict
from hug.types import text

original_wraps = functools.wraps


def wraps(function):
    """Enables building decorators around functions used for hug routes without chaninging their function signature"""
    def wrap(decorator):
        decorator = original_wraps(function)(decorator)
        if not hasattr(function, 'original'):
            decorator.original = function
        else:
            decorator.original = function.original
        return decorator
    return wrap


# def documentation(self, add_to=None):
#     """Produces general documentation for the interface"""
#     doc = OrderedDict if add_to is None else add_to
#
#     usage = self.interface.spec.__doc__
#     if 'return' in self.interface.spec.__annotations__:
#         doc['output_example'] = make_resp(json.loads(json.dumps(self.interface.spec.__annotations__['return'], default=encode_types)))
#     if usage:
#         doc['usage'] = usage
#     if getattr(self, 'requires', None):
#         doc['requires'] = [getattr(requirement, '__doc__', requirement.__name__) for requirement in self.requires]
#     doc['outputs'] = OrderedDict()
#     doc['outputs']['format'] = self.outputs.__doc__
#     doc['outputs']['content_type'] = self.outputs.content_type
#     parameters = [param for param in self.parameters if param not in (
#         'request', 'response', 'self') and not param.startswith('hug_') and not hasattr(param, 'directive')]
#     if parameters:
#         inputs = doc.setdefault('inputs', OrderedDict())
#         types = self.interface.spec.__annotations__
#         for argument in parameters:
#             kind = types.get(argument, text)
#             if getattr(kind, 'directive', None) is True:
#                 continue
#
#             input_definition = inputs.setdefault(argument, OrderedDict())
#             input_definition['type'] = kind if isinstance(kind, str) else kind.__doc__
#             default = self.defaults.get(argument, None)
#             if default is not None:
#                 input_definition['default'] = default
#
#     return doc


def gather_parameters(self, request, response, api_version=None, **input_parameters):
    """Gathers and returns all parameters that will be used for this endpoint"""
    # Mock the function parameters and add request into it
    if 'request' not in self.parameters:
        self.parameters = (*self.parameters, 'request')
    if 'request' not in self.parameters:
        self.parameters = (*self.parameters, 'response')
    input_parameters.update(request.params)
    if self.parse_body and request.content_length is not None:
        body = request.stream
        content_type, ct_params = parse_header(request.content_type)
        body_formatter = body and self.api.http.input_format(content_type)
        if body_formatter:
            body = body_formatter(body, ct_params) if ct_params else body_formatter(body)
        if 'body' in self.parameters:
            input_parameters['body'] = body
        if isinstance(body, dict):
            input_parameters.update(body)
    elif 'body' in self.parameters:
        input_parameters['body'] = None

    if 'request' in self.parameters:
        input_parameters['request'] = request
    if 'response' in self.parameters:
        input_parameters['response'] = response
    if 'api_version' in self.parameters:
        input_parameters['api_version'] = api_version
    for parameter, directive in self.directives.items():
        arguments = (self.defaults[parameter], ) if parameter in self.defaults else ()
        input_parameters[parameter] = directive(*arguments, response=response, request=request,
                                                api=self.api, api_version=api_version, interface=self)

    return input_parameters


def patch():
    psycogreen.gevent.patch_psycopg()
    functools.wraps = wraps
    # hug.interface.Interface.documentation = documentation
    hug.interface.HTTP.gather_parameters = gather_parameters



# -*- coding: utf-8 -*-

import collections


Identity = collections.namedtuple('Identity', ['code', 'error_msg', 'http_code'])


class BasicUserException(Exception):
    pass


class Unauthorized(BasicUserException):
    identity = Identity(code=0, error_msg='Unauthorized', http_code=401)


def patch_exc(exc):
    for k, v in exc.__dict__.items():
        if not k.startswith('_'):
            setattr(exc, k, type(k, (exc, ), {'identity': Identity(*v)}))
